import gymnasium
import numpy as np
from pettingzoo.utils.wrappers import BaseParallelWrapper, BaseWrapper

from supersuit.utils.frame_skip import check_transform_frameskip
from supersuit.utils.make_defaultdict import make_defaultdict
from supersuit.utils.wrapper_chooser import WrapperChooser


class frame_skip_gym(gymnasium.Wrapper):
    def __init__(self, env, num_frames):
        super().__init__(env)
        self.num_frames = check_transform_frameskip(num_frames)

    def step(self, action):
        low, high = self.num_frames
        num_skips = int(self.env.unwrapped.np_random.integers(low, high + 1))
        total_reward = 0.0

        for x in range(num_skips):
            obs, rew, term, trunc, info = super().step(action)
            total_reward += rew
            if term or trunc:
                break

        return obs, total_reward, term, trunc, info


class StepAltWrapper(BaseWrapper):
    def _modify_action(self, agent, action):
        return action

    def _modify_observation(self, agent, observation):
        return observation


class frame_skip_aec(StepAltWrapper):
    def __init__(self, env, num_frames):
        super().__init__(env)
        assert isinstance(
            num_frames, int
        ), "multi-agent frame skip only takes in an integer"
        assert num_frames > 0
        check_transform_frameskip(num_frames)
        self.num_frames = num_frames

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)
        self.agents = self.env.agents[:]
        self.terminations = make_defaultdict({agent: False for agent in self.agents})
        self.truncations = make_defaultdict({agent: False for agent in self.agents})
        self.rewards = make_defaultdict({agent: 0.0 for agent in self.agents})
        self._cumulative_rewards = make_defaultdict(
            {agent: 0.0 for agent in self.agents}
        )
        self.infos = make_defaultdict({agent: {} for agent in self.agents})
        self.skip_num = make_defaultdict({agent: 0 for agent in self.agents})
        self.old_actions = make_defaultdict({agent: None for agent in self.agents})
        self._final_observations = make_defaultdict(
            {agent: None for agent in self.agents}
        )

    def observe(self, agent):
        fin_observe = self._final_observations[agent]
        return fin_observe if fin_observe is not None else super().observe(agent)

    def step(self, action):
        self._has_updated = True

        # if agent is dead, perform a None step
        if (
            self.terminations[self.agent_selection]
            or self.truncations[self.agent_selection]
        ):
            if self.env.agents and self.agent_selection == self.env.agent_selection:
                self.env.step(None)
            self._was_dead_step(action)
            return

        cur_agent = self.agent_selection
        self._cumulative_rewards[cur_agent] = 0
        self.rewards = make_defaultdict({a: 0.0 for a in self.agents})
        self.skip_num[
            cur_agent
        ] = (
            self.num_frames
        )  # set the skip num to the param inputted in the frame_skip wrapper
        self.old_actions[cur_agent] = action

        while (
            self.old_actions[self.env.agent_selection] is not None
        ):  # this is like `for x in range(num_skips):` (L18)
            step_agent = self.env.agent_selection

            # if agent is dead, perform a None step
            if (step_agent in self.env.terminations) or (
                step_agent in self.env.truncations
            ):
                # reward = self.env.rewards[step_agent]
                # done = self.env.dones[step_agent]
                # info = self.env.infos[step_agent]
                observe, reward, termination, truncation, info = self.env.last(
                    observe=False
                )
                action = self.old_actions[step_agent]
                self.env.step(action)
                for agent in self.env.agents:
                    self.rewards[agent] += self.env.rewards[agent]
                self.infos[self.env.agent_selection] = info

                while self.env.agents and (
                    self.env.terminations[self.env.agent_selection]
                    or self.env.truncations[self.env.agent_selection]
                ):
                    dead_agent = self.env.agent_selection
                    self.terminations[dead_agent] = self.env.terminations[dead_agent]
                    self.truncations[dead_agent] = self.env.truncations[dead_agent]
                    self._final_observations[dead_agent] = self.env.observe(dead_agent)
                    self.env.step(None)
                step_agent = self.env.agent_selection

            self.skip_num[step_agent] -= 1
            if self.skip_num[step_agent] == 0:
                self.old_actions[
                    step_agent
                ] = None  # if it is time to skip, set action to None, effectively breaking the while loop

        my_agent_set = set(self.agents)
        for agent in self.env.agents:
            self.terminations[agent] = self.env.terminations[agent]
            self.truncations[agent] = self.env.truncations[agent]
            self.infos[agent] = self.env.infos[agent]
            if agent not in my_agent_set:
                self.agents.append(agent)
        self.agent_selection = self.env.agent_selection
        self._accumulate_rewards()
        self._deads_step_first()


class frame_skip_par(BaseParallelWrapper):
    def __init__(self, env, num_frames, default_action=None):
        super().__init__(env)
        self.num_frames = check_transform_frameskip(num_frames)
        self.default_action = default_action

    def step(self, action):
        action = {**action}
        low, high = self.num_frames
        num_skips = int(self.env.unwrapped.np_random.integers(low, high + 1))
        orig_agents = set(action.keys())

        total_reward = make_defaultdict({agent: 0.0 for agent in self.agents})
        total_terminations = {}
        total_truncations = {}
        total_infos = {}
        total_obs = {}

        for x in range(num_skips):
            obs, rews, term, trunc, info = super().step(action)

            for agent, rew in rews.items():
                total_reward[agent] += rew
                total_terminations[agent] = term[agent]
                total_truncations[agent] = trunc[agent]
                total_infos[agent] = info[agent]
                total_obs[agent] = obs[agent]

            for agent in self.env.agents:
                if agent not in action:
                    assert (
                        self.default_action is not None
                    ), "parallel environments that use frame_skip_v0 must provide a `default_action` argument for steps between an agent being generated and an agent taking its first step"
                    action[agent] = self.default_action

            if (
                np.fromiter(term.values(), dtype=bool)
                | np.fromiter(trunc.values(), dtype=bool)
            ).all():
                break

        # delete any values created by agents which were
        # generated and deleted before they took any actions
        final_agents = set(self.agents)
        for agent in list(total_reward):
            if agent not in final_agents and agent not in orig_agents:
                del total_reward[agent]
                del total_terminations[agent]
                del total_truncations[agent]
                del total_infos[agent]
                del total_obs[agent]

        return (
            total_obs,
            total_reward,
            total_terminations,
            total_truncations,
            total_infos,
        )


frame_skip_v0 = WrapperChooser(
    aec_wrapper=frame_skip_aec,
    gym_wrapper=frame_skip_gym,
    parallel_wrapper=frame_skip_par,
)



import functools

import gymnasium
from pettingzoo.utils import BaseParallelWrapper
from pettingzoo.utils.wrappers import OrderEnforcingWrapper as BaseWrapper

from supersuit.utils.wrapper_chooser import WrapperChooser


class shared_wrapper_aec(BaseWrapper):
    def __init__(self, env, modifier_class):
        super().__init__(env)

        self.modifier_class = modifier_class
        self.modifiers = {}
        self._cur_seed = None
        self._cur_options = None

        if hasattr(self.env, "possible_agents"):
            self.add_modifiers(self.env.possible_agents)

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        return self.modifiers[agent].modify_obs_space(self.env.observation_space(agent))

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return self.modifiers[agent].modify_action_space(self.env.action_space(agent))

    def add_modifiers(self, agents_list):
        for agent in agents_list:
            if agent not in self.modifiers:
                # populate modifier spaces
                self.modifiers[agent] = self.modifier_class()
                self.observation_space(agent)
                self.action_space(agent)
                self.modifiers[agent].reset(
                    seed=self._cur_seed, options=self._cur_options
                )

                # modifiers for each agent has a different seed
                if self._cur_seed is not None:
                    self._cur_seed += 1

    def reset(self, seed=None, options=None):
        self._cur_seed = seed
        self._cur_options = options

        for mod in self.modifiers.values():
            mod.reset(seed=seed, options=options)
        super().reset(seed=seed, options=options)

        self.add_modifiers(self.agents)
        self.modifiers[self.agent_selection].modify_obs(
            super().observe(self.agent_selection)
        )

    def step(self, action):
        mod = self.modifiers[self.agent_selection]
        action = mod.modify_action(action)
        if (
            self.terminations[self.agent_selection]
            or self.truncations[self.agent_selection]
        ):
            action = None
        super().step(action)
        self.add_modifiers(self.agents)
        self.modifiers[self.agent_selection].modify_obs(
            super().observe(self.agent_selection)
        )

    def observe(self, agent):
        return self.modifiers[agent].get_last_obs()


class shared_wrapper_parr(BaseParallelWrapper):
    def __init__(self, env, modifier_class):
        super().__init__(env)

        self.modifier_class = modifier_class
        self.modifiers = {}
        self._cur_seed = None
        self._cur_options = None

        if hasattr(self.env, "possible_agents"):
            self.add_modifiers(self.env.possible_agents)

    @functools.lru_cache(maxsize=None)
    def observation_space(self, agent):
        return self.modifiers[agent].modify_obs_space(self.env.observation_space(agent))

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        return self.modifiers[agent].modify_action_space(self.env.action_space(agent))

    def add_modifiers(self, agents_list):
        for agent in agents_list:
            if agent not in self.modifiers:
                # populate modifier spaces
                self.modifiers[agent] = self.modifier_class()
                self.observation_space(agent)
                self.action_space(agent)
                self.modifiers[agent].reset(
                    seed=self._cur_seed, options=self._cur_options
                )

                # modifiers for each agent has a different seed
                if self._cur_seed is not None:
                    self._cur_seed += 1

    def reset(self, seed=None, options=None):
        self._cur_seed = seed
        self._cur_options = options

        observations, infos = super().reset(seed=seed, options=options)
        self.add_modifiers(self.agents)
        for agent, mod in self.modifiers.items():
            mod.reset(seed=seed, options=options)
        observations = {
            agent: self.modifiers[agent].modify_obs(obs)
            for agent, obs in observations.items()
        }
        return observations, infos

    def step(self, actions):
        actions = {
            agent: self.modifiers[agent].modify_action(action)
            for agent, action in actions.items()
        }
        observations, rewards, terminations, truncations, infos = super().step(actions)
        self.add_modifiers(self.agents)
        observations = {
            agent: self.modifiers[agent].modify_obs(obs)
            for agent, obs in observations.items()
        }
        return observations, rewards, terminations, truncations, infos


class shared_wrapper_gym(gymnasium.Wrapper):
    def __init__(self, env, modifier_class):
        super().__init__(env)
        self.modifier = modifier_class()
        self.observation_space = self.modifier.modify_obs_space(self.observation_space)
        self.action_space = self.modifier.modify_action_space(self.action_space)

    def reset(self, seed=None, options=None):
        self.modifier.reset(seed=seed, options=options)
        obs, info = super().reset(seed=seed, options=options)
        obs = self.modifier.modify_obs(obs)
        return obs, info

    def step(self, action):
        obs, rew, term, trunc, info = super().step(self.modifier.modify_action(action))
        obs = self.modifier.modify_obs(obs)
        return obs, rew, term, trunc, info


shared_wrapper = WrapperChooser(
    aec_wrapper=shared_wrapper_aec,
    gym_wrapper=shared_wrapper_gym,
    parallel_wrapper=shared_wrapper_parr,
)



import copy
import multiprocessing as mp
import time
import traceback

import gymnasium.vector
import numpy as np
from gymnasium.vector.utils import (
    concatenate,
    create_empty_array,
    create_shared_memory,
    iterate,
    read_from_shared_memory,
    write_to_shared_memory,
)

from .utils.shared_array import SharedArray


def compress_info(infos):
    non_empty_infs = [(i, info) for i, info in enumerate(infos) if info]
    return non_empty_infs


def decompress_info(num_envs, idx_starts, comp_infos):
    all_info = [{}] * num_envs
    for idx_start, comp_infos in zip(idx_starts, comp_infos):
        for i, info in comp_infos:
            all_info[idx_start + i] = info
    return all_info


def write_observations(vec_env, env_start_idx, shared_obs, obs):
    obs = list(iterate(vec_env.observation_space, obs))
    for i in range(vec_env.num_envs):
        write_to_shared_memory(
            vec_env.observation_space,
            env_start_idx + i,
            obs[i],
            shared_obs,
        )


def numpy_deepcopy(buf):
    if isinstance(buf, dict):
        return {name: numpy_deepcopy(v) for name, v in buf.items()}
    elif isinstance(buf, tuple):
        return tuple(numpy_deepcopy(v) for v in buf)
    elif isinstance(buf, np.ndarray):
        return buf.copy()
    else:
        raise ValueError("numpy_deepcopy ")


def async_loop(
    vec_env_constr, inpt_p, pipe, shared_obs, shared_rews, shared_terms, shared_truncs
):
    inpt_p.close()
    try:
        vec_env = vec_env_constr()

        pipe.send(vec_env.num_envs)
        env_start_idx = pipe.recv()
        env_end_idx = env_start_idx + vec_env.num_envs
        while True:
            instr = pipe.recv()
            comp_infos = []

            if instr == "close":
                vec_env.close()

            elif isinstance(instr, tuple):
                name, data = instr

                if name == "reset":
                    observations, infos = vec_env.reset(seed=data[0], options=data[1])
                    comp_infos = compress_info(infos)

                    write_observations(vec_env, env_start_idx, shared_obs, observations)
                    shared_terms.np_arr[env_start_idx:env_end_idx] = False
                    shared_truncs.np_arr[env_start_idx:env_end_idx] = False
                    shared_rews.np_arr[env_start_idx:env_end_idx] = 0.0

                elif name == "step":
                    actions = data
                    actions = concatenate(
                        vec_env.action_space,
                        actions,
                        create_empty_array(vec_env.action_space, n=len(actions)),
                    )
                    observations, rewards, terms, truncs, infos = vec_env.step(actions)
                    write_observations(vec_env, env_start_idx, shared_obs, observations)
                    shared_terms.np_arr[env_start_idx:env_end_idx] = terms
                    shared_truncs.np_arr[env_start_idx:env_end_idx] = truncs
                    shared_rews.np_arr[env_start_idx:env_end_idx] = rewards
                    comp_infos = compress_info(infos)

                elif name == "env_is_wrapped":
                    comp_infos = vec_env.env_is_wrapped(data)

                else:
                    raise AssertionError("bad tuple instruction name: " + name)
            elif instr == "render":
                render_result = vec_env.render()
                if vec_env.render_mode == "rgb_array":
                    comp_infos = render_result
            elif instr == "terminate":
                return
            else:
                raise AssertionError("bad instruction: " + instr)
            pipe.send(comp_infos)
    except BaseException as e:
        tb = traceback.format_exc()
        pipe.send((e, tb))


class ProcConcatVec(gymnasium.vector.VectorEnv):
    def __init__(
        self, vec_env_constrs, observation_space, action_space, tot_num_envs, metadata
    ):
        self.observation_space = observation_space
        self.action_space = action_space
        self.num_envs = num_envs = tot_num_envs
        self.metadata = metadata

        self.shared_obs = create_shared_memory(self.observation_space, n=self.num_envs)
        self.shared_act = create_shared_memory(self.action_space, n=self.num_envs)
        self.shared_rews = SharedArray((num_envs,), dtype=np.float32)
        self.shared_terms = SharedArray((num_envs,), dtype=np.uint8)
        self.shared_truncs = SharedArray((num_envs,), dtype=np.uint8)

        self.observations_buffers = read_from_shared_memory(
            self.observation_space, self.shared_obs, n=self.num_envs
        )

        self.graceful_shutdown_timeout = 10

        pipes = []
        procs = []
        for constr in vec_env_constrs:
            inpt, outpt = mp.Pipe()
            constr = gymnasium.vector.async_vector_env.CloudpickleWrapper(constr)
            proc = mp.Process(
                target=async_loop,
                args=(
                    constr,
                    inpt,
                    outpt,
                    self.shared_obs,
                    self.shared_rews,
                    self.shared_terms,
                    self.shared_truncs,
                ),
            )
            proc.start()
            outpt.close()
            pipes.append(inpt)
            procs.append(proc)

        self.pipes = pipes
        self.procs = procs

        num_envs = 0
        env_nums = self._receive_info()
        idx_starts = []
        for pipe, cnum_env in zip(self.pipes, env_nums):
            cur_env_idx = num_envs
            num_envs += cnum_env
            pipe.send(cur_env_idx)
            idx_starts.append(cur_env_idx)
        idx_starts.append(num_envs)

        assert num_envs == tot_num_envs
        self.idx_starts = idx_starts

    def reset(self, seed=None, options=None):
        for i, pipe in enumerate(self.pipes):
            if seed is not None:
                pipe.send(("reset", (seed + i, options)))
            else:
                pipe.send(("reset", (seed, options)))

        info = self._receive_info()

        return numpy_deepcopy(self.observations_buffers), copy.deepcopy(info)

    def step_async(self, actions):
        actions = list(iterate(self.action_space, actions))
        for i, pipe in enumerate(self.pipes):
            start, end = self.idx_starts[i : i + 2]
            pipe.send(("step", actions[start:end]))

    def _receive_info(self):
        all_data = []
        for cin in self.pipes:
            data = cin.recv()
            if isinstance(data, tuple):
                e, tb = data
                print(tb)
                raise e
            all_data.append(data)
        return all_data

    def step_wait(self):
        compressed_infos = self._receive_info()
        infos = decompress_info(self.num_envs, self.idx_starts, compressed_infos)
        rewards = self.shared_rews.np_arr
        terms = self.shared_terms.np_arr
        truncs = self.shared_truncs.np_arr
        return (
            numpy_deepcopy(self.observations_buffers),
            rewards.copy(),
            terms.astype(bool).copy(),
            truncs.astype(bool).copy(),
            copy.deepcopy(infos),
        )

    def step(self, actions):
        self.step_async(actions)
        return self.step_wait()

    def __del__(self):
        self.close()

    def render(self):
        self.pipes[0].send("render")
        render_result = self.pipes[0].recv()

        if isinstance(render_result, tuple):
            e, tb = render_result
            print(tb)
            raise e

        return render_result

    def close(self):
        try:
            for pipe, proc in zip(self.pipes, self.procs):
                if proc.is_alive():
                    pipe.send("close")
        except OSError:
            pass
        else:
            deadline = (
                None
                if self.graceful_shutdown_timeout is None
                else time.monotonic() + self.graceful_shutdown_timeout
            )
            for proc in self.procs:
                timeout = None if deadline is None else deadline - time.monotonic()
                if timeout is not None and timeout <= 0:
                    break
                proc.join(timeout)
        for pipe, proc in zip(self.pipes, self.procs):
            if proc.is_alive():
                proc.kill()
            pipe.close()

    def env_is_wrapped(self, wrapper_class, indices=None):
        for i, pipe in enumerate(self.pipes):
            pipe.send(("env_is_wrapped", wrapper_class))

        results = self._receive_info()
        return sum(results, [])



import gymnasium

from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper


def sticky_actions_v0(env, repeat_action_probability):
    assert 0 <= repeat_action_probability < 1

    class StickyActionsModifier(BaseModifier):
        def __init__(self):
            super().__init__()

        def reset(self, seed=None, options=None):
            self.np_random, _ = gymnasium.utils.seeding.np_random(seed)
            self.old_action = None

        def modify_action(self, action):
            if (
                self.old_action is not None
                and self.np_random.uniform() < repeat_action_probability
            ):
                action = self.old_action
            self.old_action = action
            return action

    return shared_wrapper(env, StickyActionsModifier)



import warnings
from typing import Any, List, Optional

import numpy as np
from stable_baselines3.common.vec_env import VecEnvWrapper
from stable_baselines3.common.vec_env.base_vec_env import VecEnvIndices


class SB3VecEnvWrapper(VecEnvWrapper):
    def __init__(self, venv):
        self.venv = venv
        self.num_envs = venv.num_envs
        self.observation_space = venv.observation_space
        self.action_space = venv.action_space
        self.reset_infos = []

    def reset(self, seed=None, options=None):
        if seed is not None:
            self.seed(seed=seed)
        # Note: SB3's vector envs return only observations on reset, and store infos in `self.reset_infos`
        observations, self.reset_infos = self.venv.reset()
        return observations

    def step_wait(self):
        observations, rewards, terminations, truncations, infos = self.venv.step_wait()
        # Note: SB3 expects dones to be an np.array
        dones = np.array(
            [terminations[i] or truncations[i] for i in range(len(terminations))]
        )
        return observations, rewards, dones, infos

    def env_is_wrapped(self, wrapper_class, indices=None):
        # ignores indices
        return self.venv.env_is_wrapped(wrapper_class)

    def getattr_recursive(self, name):
        raise AttributeError(name)

    def getattr_depth_check(self, name, already_found):
        return None

    def get_attr(self, attr_name: str, indices: VecEnvIndices = None) -> List[Any]:
        attr = self.venv.get_attr(attr_name)
        # Note: SB3 expects render_mode to be returned as an array, with values for each env
        if attr_name == "render_mode":
            return [attr for _ in range(self.num_envs)]
        else:
            return attr

    def render(self, mode: Optional[str] = None) -> Optional[np.ndarray]:
        warnings.warn(
            "PettingZoo environments do not take the `render(mode)` argument, to change rendering mode, re-initialize the environment using the `render_mode` argument."
        )
        return self.venv.render()



import warnings

import gymnasium
import numpy as np

from supersuit.lambda_wrappers import action_lambda_v1

from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper


def nan_random_v0(env):
    class NanRandomModifier(BaseModifier):
        def __init__(self):
            super().__init__()

        def reset(self, seed=None, options=None):
            self.np_random, seed = gymnasium.utils.seeding.np_random(seed)

            return super().reset(seed, options=options)

        def modify_action(self, action):
            if action is not None and np.isnan(action).any():
                obs = self.cur_obs
                if isinstance(obs, dict) and "action mask" in obs:
                    warnings.warn(
                        "[WARNING]: Step received an NaN action {}. Environment is {}. Taking a random action from 'action mask'.".format(
                            action, self
                        )
                    )
                    action = self.np_random.choice(np.flatnonzero(obs["action_mask"]))
                else:
                    warnings.warn(
                        "[WARNING]: Step received an NaN action {}. Environment is {}. Taking a random action.".format(
                            action, self
                        )
                    )
                    action = self.action_space.sample()

            return action

    return shared_wrapper(env, NanRandomModifier)


def nan_noop_v0(env, no_op_action):
    def on_action(action, action_space):
        if action is None:
            warnings.warn(
                "[WARNING]: Step received an None action {}. Environment is {}. Taking no operation action.".format(
                    action, env
                )
            )
            return None
        if np.isnan(action).any():
            warnings.warn(
                "[WARNING]: Step received an NaN action {}. Environment is {}. Taking no operation action.".format(
                    action, env
                )
            )
            return no_op_action
        return action

    return action_lambda_v1(env, on_action, lambda act_space: act_space)


def nan_zeros_v0(env):
    def on_action(action, action_space):
        if action is None:
            warnings.warn(
                "[WARNING]: Step received an None action {}. Environment is {}. Taking the all zeroes action.".format(
                    action, env
                )
            )
            return None
        if np.isnan(action).any():
            warnings.warn(
                "[WARNING]: Step received an NaN action {}. Environment is {}. Taking the all zeroes action.".format(
                    action, env
                )
            )
            return np.zeros_like(action)
        return action

    return action_lambda_v1(env, on_action, lambda act_space: act_space)



from gymnasium.spaces import Box, Discrete

from supersuit.utils.frame_stack import stack_init, stack_obs, stack_obs_space

from .utils.base_modifier import BaseModifier
from .utils.shared_wrapper_util import shared_wrapper


def frame_stack_v1(env, stack_size=4, stack_dim=-1):
    assert isinstance(stack_size, int), "stack size of frame_stack must be an int"

    class FrameStackModifier(BaseModifier):
        def modify_obs_space(self, obs_space):
            if isinstance(obs_space, Box):
                assert (
                    1 <= len(obs_space.shape) <= 3
                ), "frame_stack only works for 1, 2 or 3 dimensional observations"
            elif isinstance(obs_space, Discrete):
                pass
            else:
                assert (
                    False
                ), "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(
                    obs_space
                )

            self.old_obs_space = obs_space
            self.observation_space = stack_obs_space(obs_space, stack_size, stack_dim)
            return self.observation_space

        def reset(self, seed=None, options=None):
            self.stack = stack_init(self.old_obs_space, stack_size, stack_dim)

        def modify_obs(self, obs):
            self.stack = stack_obs(
                self.stack, obs, self.old_obs_space, stack_size, stack_dim
            )

            return self.stack

        def get_last_obs(self):
            return self.stack

    return shared_wrapper(env, FrameStackModifier)


def frame_stack_v2(env, stack_size=4, stack_dim=-1):
    assert isinstance(stack_size, int), "stack size of frame_stack must be an int"
    assert f"stack_dim should be 0 or -1, not {stack_dim}"

    class FrameStackModifier(BaseModifier):
        def modify_obs_space(self, obs_space):
            if isinstance(obs_space, Box):
                assert (
                    1 <= len(obs_space.shape) <= 3
                ), "frame_stack only works for 1, 2 or 3 dimensional observations"
            elif isinstance(obs_space, Discrete):
                pass
            else:
                assert (
                    False
                ), "Stacking is currently only allowed for Box and Discrete observation spaces. The given observation space is {}".format(
                    obs_space
                )

            self.old_obs_space = obs_space
            self.observation_space = stack_obs_space(obs_space, stack_size, stack_dim)
            return self.observation_space

        def reset(self, seed=None, options=None):
            self.stack = stack_init(self.old_obs_space, stack_size, stack_dim)
            self.reset_flag = True

        def modify_obs(self, obs):
            if self.reset_flag:
                for _ in range(stack_size):
                    self.stack = stack_obs(
                        self.stack, obs, self.old_obs_space, stack_size, stack_dim
                    )
                self.reset_flag = False
            else:
                self.stack = stack_obs(
                    self.stack, obs, self.old_obs_space, stack_size, stack_dim
                )

            return self.stack

        def get_last_obs(self):
            return self.stack

    return shared_wrapper(env, FrameStackModifier)



from collections import deque

import numpy as np


class Delayer:
    def __init__(self, obs_space, delay):
        self.delay = delay
        self.obs_queue = deque()
        self.obs_space = obs_space

    def add(self, in_obs):
        self.obs_queue.append(in_obs)
        if len(self.obs_queue) > self.delay:
            return self.obs_queue.popleft()
        else:
            if isinstance(in_obs, np.ndarray):
                return np.zeros_like(in_obs)
            elif (
                isinstance(in_obs, dict)
                and "observation" in in_obs.keys()
                and "action_mask" in in_obs.keys()
            ):
                return {
                    "observation": np.zeros_like(in_obs["observation"]),
                    "action_mask": np.ones_like(in_obs["action_mask"]),
                }
            else:
                raise TypeError(
                    "Observation must be of type np.ndarray or dictionary with keys 'observation' and 'action_mask'"
                )



import gymnasium.vector
import numpy as np
from gymnasium.spaces import Discrete
from gymnasium.vector.utils import concatenate, create_empty_array, iterate

from .single_vec_env import SingleVecEnv


def transpose(ll):
    return [[ll[i][j] for i in range(len(ll))] for j in range(len(ll[0]))]


@iterate.register(Discrete)
def iterate_discrete(space, items):
    try:
        return iter(items)
    except TypeError:
        raise TypeError(f"Unable to iterate over the following elements: {items}")


class ConcatVecEnv(gymnasium.vector.VectorEnv):
    def __init__(self, vec_env_fns, obs_space=None, act_space=None):
        self.vec_envs = vec_envs = [vec_env_fn() for vec_env_fn in vec_env_fns]
        for i in range(len(vec_envs)):
            if not hasattr(vec_envs[i], "num_envs"):
                vec_envs[i] = SingleVecEnv([lambda: vec_envs[i]])
        self.metadata = self.vec_envs[0].metadata
        self.render_mode = self.vec_envs[0].render_mode
        self.observation_space = vec_envs[0].observation_space
        self.action_space = vec_envs[0].action_space
        tot_num_envs = sum(env.num_envs for env in vec_envs)
        self.num_envs = tot_num_envs

    def reset(self, seed=None, options=None):
        _res_obs = []
        _res_infos = []

        if seed is not None:
            for i in range(len(self.vec_envs)):
                _obs, _info = self.vec_envs[i].reset(seed=seed + i, options=options)
                _res_obs.append(_obs)
                _res_infos.append(_info)
        else:
            for i in range(len(self.vec_envs)):
                _obs, _info = self.vec_envs[i].reset(options=options)
                _res_obs.append(_obs)
                _res_infos.append(_info)

        # flatten infos (also done in step function)
        flattened_infos = [info for sublist in _res_infos for info in sublist]

        return self.concat_obs(_res_obs), flattened_infos

    def concat_obs(self, observations):
        return concatenate(
            self.observation_space,
            [
                item
                for obs in observations
                for item in iterate(self.observation_space, obs)
            ],
            create_empty_array(self.observation_space, n=self.num_envs),
        )

    def concatenate_actions(self, actions, n_actions):
        return concatenate(
            self.action_space,
            actions,
            create_empty_array(self.action_space, n=n_actions),
        )

    def step_async(self, actions):
        self._saved_actions = actions

    def step_wait(self):
        return self.step(self._saved_actions)

    def step(self, actions):
        data = []
        idx = 0
        actions = list(iterate(self.action_space, actions))
        for venv in self.vec_envs:
            data.append(
                venv.step(
                    self.concatenate_actions(
                        actions[idx : idx + venv.num_envs], venv.num_envs
                    )
                )
            )
            idx += venv.num_envs
        observations, rewards, terminations, truncations, infos = transpose(data)
        observations = self.concat_obs(observations)
        rewards = np.concatenate(rewards, axis=0)
        terminations = np.concatenate(terminations, axis=0)
        truncations = np.concatenate(truncations, axis=0)
        infos = [
            info for sublist in infos for info in sublist
        ]  # flatten infos from nested lists
        return observations, rewards, terminations, truncations, infos

    def render(self):
        return self.vec_envs[0].render()

    def close(self):
        for vec_env in self.vec_envs:
            vec_env.close()

    def env_is_wrapped(self, wrapper_class):
        return sum(
            [sub_venv.env_is_wrapped(wrapper_class) for sub_venv in self.vec_envs], []
        )



import functools

import gymnasium
from gymnasium.spaces import Space

from supersuit.utils.base_aec_wrapper import BaseWrapper
from supersuit.utils.wrapper_chooser import WrapperChooser


class aec_action_lambda(BaseWrapper):
    def __init__(self, env, change_action_fn, change_space_fn):
        assert callable(
            change_action_fn
        ), f"change_action_fn needs to be a function. It is {change_action_fn}"
        assert callable(
            change_space_fn
        ), f"change_space_fn needs to be a function. It is {change_space_fn}"

        self.change_action_fn = change_action_fn
        self.change_space_fn = change_space_fn

        super().__init__(env)
        if hasattr(self, "possible_agents"):
            for agent in self.possible_agents:
                # call any validation logic in this function
                self.action_space(agent)

    def _modify_observation(self, agent, observation):
        return observation

    @functools.lru_cache(maxsize=None)
    def action_space(self, agent):
        old_act_space = self.env.action_space(agent)
        try:
            return self.change_space_fn(old_act_space, agent)
        except TypeError:
            return self.change_space_fn(old_act_space)

    def _modify_action(self, agent, action):
        old_act_space = self.env.action_space(agent)
        try:
            return self.change_action_fn(action, old_act_space, agent)
        except TypeError:
            return self.change_action_fn(action, old_act_space)


class gym_action_lambda(gymnasium.Wrapper):
    def __init__(self, env, change_action_fn, change_space_fn):
        assert callable(
            change_action_fn
        ), f"change_action_fn needs to be a function. It is {change_action_fn}"
        assert callable(
            change_space_fn
        ), f"change_space_fn needs to be a function. It is {change_space_fn}"
        self.change_action_fn = change_action_fn
        self.change_space_fn = change_space_fn

        super().__init__(env)
        self._modify_spaces()

    def _modify_spaces(self):
        new_space = self.change_space_fn(self.action_space)
        assert isinstance(
            new_space, Space
        ), "output of change_space_fn argument to action_lambda_wrapper must be a gymnasium space"
        self.action_space = new_space

    def _modify_action(self, action):
        return self.change_action_fn(action, self.env.action_space)

    def step(self, action):
        return super().step(self._modify_action(action))


action_lambda_v1 = WrapperChooser(
    aec_wrapper=aec_action_lambda, gym_wrapper=gym_action_lambda
)



import gymnasium
from pettingzoo.utils import BaseWrapper as PettingzooWrap

from supersuit.utils.make_defaultdict import make_defaultdict
from supersuit.utils.wrapper_chooser import WrapperChooser


class aec_reward_lambda(PettingzooWrap):
    def __init__(self, env, change_reward_fn):
        assert callable(
            change_reward_fn
        ), f"change_reward_fn needs to be a function. It is {change_reward_fn}"
        self._change_reward_fn = change_reward_fn

        super().__init__(env)

    def _check_wrapper_params(self):
        pass

    def _modify_spaces(self):
        pass

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)
        self.rewards = {
            agent: self._change_reward_fn(reward)
            for agent, reward in self.env.rewards.items()  # you don't want to unwrap here, because another reward wrapper might have been applied
        }
        self.__cumulative_rewards = make_defaultdict({a: 0 for a in self.agents})
        self._accumulate_rewards()

    def step(self, action):
        agent = self.env.agent_selection
        super().step(action)
        self.rewards = {
            agent: self._change_reward_fn(reward)
            for agent, reward in self.env.rewards.items()  # you don't want to unwrap here, because another reward wrapper might have been applied
        }
        self.__cumulative_rewards[agent] = 0
        self._cumulative_rewards = self.__cumulative_rewards
        self._accumulate_rewards()


class gym_reward_lambda(gymnasium.Wrapper):
    def __init__(self, env, change_reward_fn):
        assert callable(
            change_reward_fn
        ), f"change_reward_fn needs to be a function. It is {change_reward_fn}"
        self._change_reward_fn = change_reward_fn

        super().__init__(env)

    def step(self, action):
        obs, rew, termination, truncation, info = super().step(action)
        return obs, self._change_reward_fn(rew), termination, truncation, info


reward_lambda_v0 = WrapperChooser(
    aec_wrapper=aec_reward_lambda, gym_wrapper=gym_reward_lambda
)



from pettingzoo.utils.env import AECEnv, ParallelEnv

from supersuit import observation_lambda_v0
from supersuit.utils import agent_indicator as agent_ider


def agent_indicator_v0(env, type_only=False):
    assert isinstance(env, AECEnv) or isinstance(
        env, ParallelEnv
    ), "agent_indicator_v0 only accepts an AECEnv or ParallelEnv"
    assert hasattr(
        env, "possible_agents"
    ), "environment passed to agent indicator wrapper must have the possible_agents attribute."

    indicator_map = agent_ider.get_indicator_map(env.possible_agents, type_only)
    num_indicators = len(set(indicator_map.values()))

    obs_spaces = [env.observation_space(agent) for agent in env.possible_agents]
    agent_ider.check_params(obs_spaces)

    return observation_lambda_v0(
        env,
        lambda obs, obs_space, agent: agent_ider.change_observation(
            obs,
            obs_space,
            (indicator_map[agent], num_indicators),
        ),
        lambda obs_space: agent_ider.change_obs_space(obs_space, num_indicators),
    )



import pytest

import supersuit


def test_bad_import():
    with pytest.raises(supersuit.DeprecatedWrapper):
        from supersuit import action_lambda_v0  # noqa: F401
    with pytest.raises(supersuit.DeprecatedWrapper):
        supersuit.action_lambda_v0



import numpy as np

from . import convert_box


def check_param(obs_space, shape):
    assert isinstance(shape, tuple), f"shape must be tuple. It is {shape}"
    assert all(
        isinstance(el, int) for el in shape
    ), f"shape must be tuple of ints, is: {shape}"
    assert np.prod(shape) == np.prod(
        obs_space.shape
    ), "new shape {} must have as many elements as original shape {}".format(
        shape, obs_space.shape
    )


def change_obs_space(obs_space, param):
    return convert_box(lambda obs: change_observation(obs, obs_space, param), obs_space)


def change_observation(obs, obs_space, shape):
    obs = obs.reshape(shape)
    return obs



import cloudpickle
from pettingzoo import AECEnv

from .async_vector_env import AsyncAECVectorEnv
from .vector_env import SyncAECVectorEnv


def vectorize_aec_env_v0(aec_env, num_envs, num_cpus=0):
    assert isinstance(
        aec_env, AECEnv
    ), "pettingzoo_env_to_vec_env takes in a pettingzoo AECEnv."
    assert hasattr(
        aec_env, "possible_agents"
    ), "environment passed to vectorize_aec_env must have possible_agents attribute."

    def env_fn():
        return cloudpickle.loads(cloudpickle.dumps(aec_env))

    env_list = [env_fn] * num_envs

    if num_cpus == 0 or num_cpus == 1:
        return SyncAECVectorEnv(env_list)
    else:
        return AsyncAECVectorEnv(env_list, num_cpus)



import gymnasium.vector
import numpy as np
from gymnasium.vector.utils import concatenate, create_empty_array, iterate


class MarkovVectorEnv(gymnasium.vector.VectorEnv):
    def __init__(self, par_env, black_death=False):
        """
        parameters:
            - par_env: the pettingzoo Parallel environment that will be converted to a gymnasium vector environment
            - black_death: whether to give zero valued observations and 0 rewards when an agent is done, allowing for environments with multiple numbers of agents.
                            Is equivalent to adding the black death wrapper, but somewhat more efficient.

        The resulting object will be a valid vector environment that has a num_envs
        parameter equal to the max number of agents, will return an array of observations,
        rewards, dones, etc, and will reset environment automatically when it finishes
        """
        self.par_env = par_env
        self.metadata = par_env.metadata
        self.render_mode = par_env.unwrapped.render_mode
        self.observation_space = par_env.observation_space(par_env.possible_agents[0])
        self.action_space = par_env.action_space(par_env.possible_agents[0])
        assert all(
            self.observation_space == par_env.observation_space(agent)
            for agent in par_env.possible_agents
        ), "observation spaces not consistent. Perhaps you should wrap with `supersuit.multiagent_wrappers.pad_observations_v0`?"
        assert all(
            self.action_space == par_env.action_space(agent)
            for agent in par_env.possible_agents
        ), "action spaces not consistent. Perhaps you should wrap with `supersuit.multiagent_wrappers.pad_action_space_v0`?"
        self.num_envs = len(par_env.possible_agents)
        self.black_death = black_death

    def concat_obs(self, obs_dict):
        obs_list = []
        for i, agent in enumerate(self.par_env.possible_agents):
            if agent not in obs_dict:
                raise AssertionError(
                    "environment has agent death. Not allowed for pettingzoo_env_to_vec_env_v1 unless black_death is True"
                )
            obs_list.append(obs_dict[agent])

        return concatenate(
            self.observation_space,
            obs_list,
            create_empty_array(self.observation_space, self.num_envs),
        )

    def step_async(self, actions):
        self._saved_actions = actions

    def step_wait(self):
        return self.step(self._saved_actions)

    def reset(self, seed=None, options=None):
        # TODO: should this be changed to infos?
        _observations, infos = self.par_env.reset(seed=seed, options=options)
        observations = self.concat_obs(_observations)
        infs = [infos.get(agent, {}) for agent in self.par_env.possible_agents]
        return observations, infs

    def step(self, actions):
        actions = list(iterate(self.action_space, actions))
        agent_set = set(self.par_env.agents)
        act_dict = {
            agent: actions[i]
            for i, agent in enumerate(self.par_env.possible_agents)
            if agent in agent_set
        }
        observations, rewards, terms, truncs, infos = self.par_env.step(act_dict)

        # adds last observation to info where user can get it
        terminations = np.fromiter(terms.values(), dtype=bool)
        truncations = np.fromiter(truncs.values(), dtype=bool)
        env_done = (terminations | truncations).all()
        if env_done:
            for agent, obs in observations.items():
                infos[agent]["terminal_observation"] = obs

        rews = np.array(
            [rewards.get(agent, 0) for agent in self.par_env.possible_agents],
            dtype=np.float32,
        )
        tms = np.array(
            [terms.get(agent, False) for agent in self.par_env.possible_agents],
            dtype=np.uint8,
        )
        tcs = np.array(
            [truncs.get(agent, False) for agent in self.par_env.possible_agents],
            dtype=np.uint8,
        )
        infs = [infos.get(agent, {}) for agent in self.par_env.possible_agents]

        if env_done:
            observations, reset_infs = self.reset()
        else:
            observations = self.concat_obs(observations)
            # empty infos for reset infs
            reset_infs = [{} for _ in range(len(self.par_env.possible_agents))]
        # combine standard infos and reset infos
        infs = [{**inf, **reset_inf} for inf, reset_inf in zip(infs, reset_infs)]

        assert (
            self.black_death or self.par_env.agents == self.par_env.possible_agents
        ), "MarkovVectorEnv does not support environments with varying numbers of active agents unless black_death is set to True"
        return observations, rews, tms, tcs, infs

    def render(self):
        return self.par_env.render()

    def close(self):
        return self.par_env.close()

    def env_is_wrapped(self, wrapper_class):
        """
        env_is_wrapped only suppors vector and gymnasium environments
        currently, not pettingzoo environments
        """
        return [False] * self.num_envs



import warnings
from collections import defaultdict


def make_defaultdict(d):
    try:
        dd = defaultdict(type(next(iter(d.values()))))
        for k, v in d.items():
            dd[k] = v
        return dd
    except StopIteration:
        warnings.warn("No agents left in the environment!")
        return {}



import numpy as np
from gymnasium.spaces import Box


def check_param(obs_space, min_max_pair):
    assert np.dtype(obs_space.dtype) == np.dtype("float32") or np.dtype(
        obs_space.dtype
    ) == np.dtype("float64")
    assert (
        isinstance(min_max_pair, tuple) and len(min_max_pair) == 2
    ), f"range_scale must be tuple of size 2. It is {min_max_pair}"
    try:
        min_res = float(min_max_pair[0])
        max_res = float(min_max_pair[1])
    except ValueError:
        assert False, "normalize_obs inputs must be numbers. They are {}".format(
            min_max_pair
        )
    assert (
        max_res > min_res
    ), "maximum must be greater than minimum value in normalize_obs"
    assert np.all(np.isfinite(obs_space.low)) and np.all(
        np.isfinite(obs_space.high)
    ), "Box observation_space of environment has infinite bounds! Only environments with finite bounds can be passed to normalize_obs_v0"


def change_obs_space(obs_space, min_max_pair):
    min = np.float64(min_max_pair[0]).astype(obs_space.dtype)
    max = np.float64(min_max_pair[1]).astype(obs_space.dtype)
    return Box(low=min, high=max, shape=obs_space.shape, dtype=obs_space.dtype)


def change_observation(obs, obs_space, min_max_pair):
    min_res, max_res = (float(x) for x in min_max_pair)
    old_size = obs_space.high - obs_space.low
    new_size = max_res - min_res
    result = (obs - obs_space.low) / old_size * new_size + min_res
    return result
